"""EO4NOWCAST REAL TIME FLOOD PREDICTION DAG"""
from typing import List
from flood_risk.config import STATION_PREFIX_FILE, FLOODRISK_MODEL_NAMES, HOURLY_PLUVIO_FLOOD_RISK_FOLDER, AIRFLOW_NDMI_FOLDER, AIRFLOW_MODELS_FOLDER, AIRFLOW_CUMULATED_PLUVIO_FOLDER, AIRFLOW_PREDICTED_NDMI_FOLDER

from airflow.decorators import dag, task
from airflow.operators.python import get_current_context
from airflow.operators.email import EmailOperator
from airflow.operators.bash import BashOperator
from datetime import datetime

# def catch_error(context):
#     """Sends an email in case of error"""
#     task_id = context['ti'].task_id
#     dag_name = context['dag'].dag_id
#     error_mail = EmailOperator(
#         task_id='error_mail',
#         to='team.turing@stamtech.com',
#         subject=f'Airflow {dag_name} FAIL',
#         html_content=f'Task {task_id} failed',
#         mime_charset='utf-8')
#     error_mail.execute({})  # no need to return, just execute


@dag(
    start_date=datetime(2021, 1, 1, 0, 0, 0),
    schedule_interval="*/5 * * * *",
    default_args={
        # "on_failure_callback": catch_error,
        "retries": 0,
        "retry_delay": 30,
        "retry_exponential_backoff": True
    },
    params={},
    catchup=False
)
def Flood_Risk():
    """Manage data for flood prediction, predict flood risk level with a CNN model"""
    @task
    def get_cumulated_pluvio(now: datetime) -> dict:
        print("Started pluvio retriever")
        from flood_risk.extract import get_pluvio_from_ftp_server
        now = now.replace(minute=0, hour=now.hour-1)

        print("now", now)
        local_paths = get_pluvio_from_ftp_server(now)
        current_date = now.strftime("%Y%m%dT%H%M")
      
        result = {
            'paths': local_paths,
            'current_date': current_date
        }
        return result        

    @task
    def compute_partial_pluvio(latest_pluvio_local_paths: List[str], current_date: str) -> str:
        print("Computing latest hourly pluvio image")
        from flood_risk.transform import create_partial_pluvio
        result = create_partial_pluvio(latest_pluvio_local_paths, current_date)

        delete_ftp_flood_risk_output_files = BashOperator(
            task_id="delete_ftp_flood_risk_output_files",
            run_as_user='artys',
            bash_command=('sudo rm -r /home/artys/ftp_flood_risk_output/*; ' )
        )
        delete_ftp_flood_risk_output_files.execute({}) #clean ftp_output_flood_risk files
        return result
    
    @task(task_id="fetch_latest_predicted_ndmi")
    def fetch_latest_predicted_ndmi() -> str:
        """
        Retrieve latest predicted NDMI searching both in 
        the local filesystem and in the HuggingFace repository
        """
        # check if there is already a latest ndmi in main folder
        from flood_risk.extract import get_latest_predicted_ndmi
        predicted_ndmi_filepath = get_latest_predicted_ndmi()
        #return both filesystem local path of the predicted NDMI 
        return predicted_ndmi_filepath
    
    @task(task_id="fetch_mlm_floodrisk_alerts")
    def fetch_mlm_floodrisk_alerts() -> dict:
        """
        Retrieve the models for predicting the flood risk of all stations 
        """
        # check if the models are already available in the local filesystem
        from flood_risk.extract import get_mlm_models
        models_registry = get_mlm_models()
        return models_registry 
    @task
    def create_new_predicted_flood_risk(hourly_pluvio_filepath, ndmi_local_filepath, models_registry, now):
        from flood_risk.transform import run_mlm_flood_risk
        from flood_risk.load import save_risk_value_on_hf
        current_date = now.strftime("%Y%m%dT%H%M")
        for m in FLOODRISK_MODEL_NAMES:
            station = m.split("-")[1]
            risk_value = run_mlm_flood_risk(models_registry[station], hourly_pluvio_filepath, ndmi_local_filepath)
            save_risk_value_on_hf(current_date, risk_value, STATION_PREFIX_FILE+station)

        delete_results_station_files = BashOperator(
            task_id="delete_results_station_files",
            run_as_user='artys',
            bash_command=('sudo rm -r /home/airflow/result_station/*; ' )
        )
        delete_results_station_files.execute({}) #clean ftp_output_flood_risk files

    now = datetime.now()
    ndmi_local_filepath = fetch_latest_predicted_ndmi()
    models_registry = fetch_mlm_floodrisk_alerts()

    result = get_cumulated_pluvio(now)

    hourly_pluvio_filepath = compute_partial_pluvio(result['paths'], result['current_date'])
    
    create_new_predicted_flood_risk(hourly_pluvio_filepath, ndmi_local_filepath, models_registry, now)


Flood_Risk()
