"""EO4NOWCAST REAL TIME NDMI PREDICTION DAG"""
from ndmi_prediction.config import AIRFLOW_NDMI_FOLDER, ARTYS_LATEST_NDMI_FOLDER, AIRFLOW_MODELS_FOLDER, NDMI_MODEL_NAME, AIRFLOW_CUMULATED_PLUVIO_FOLDER, AIRFLOW_PREDICTED_NDMI_FOLDER
from datetime import datetime, timedelta
from typing import List, Tuple

from airflow.decorators import dag, task
from airflow.operators.python import get_current_context
from airflow.operators.email import EmailOperator
from airflow.operators.bash import BashOperator
from airflow.models import Variable
from airflow.utils.trigger_rule import TriggerRule

# def catch_error(context):
#     """Sends an email in case of error"""
#     task_id = context['ti'].task_id
#     dag_name = context['dag'].dag_id
#     error_mail = EmailOperator(
#         task_id='error_mail',
#         to='team.turing@stamtech.com',
#         subject=f'Airflow {dag_name} FAIL',
#         html_content=f'Task {task_id} failed',
#         mime_charset='utf-8')
#     error_mail.execute({})  # no need to return, just execute


@dag(
    start_date=datetime(2021, 1, 1, 0, 0, 0),
    schedule_interval="0 * * * *",
    default_args={
        # "on_failure_callback": catch_error,
        "retries": 0,
        "retry_delay": 30,
        "retry_exponential_backoff": True
    },
    params={},
    catchup=False
)
def NDMIPrediction():
    """Manage data for flood prediction, retrieve new data from Artys, create data for ML models, predict new NDMI by
    UNET model"""

    @task
    def artys_ndmi_retriever( task_id = "ndmi_retriever"):
        """Retrieve new ndmi by artys_ndmi_retriever"""
        end_date = datetime.now()
        start_date = end_date - timedelta(hours=1, minutes=0)
        start_date = start_date.strftime('%Y-%m-%dT%H:%M')   # "2022-08-22T14:00"
        end_date = end_date.strftime('%Y-%m-%dT%H:%M')   # "2022-08-24T15:00"  
        print("start_date ", start_date)
        print("end_date ", end_date)
        
        artys_ndmi_retriever = BashOperator(
            task_id="artys_ndmi_retriever",
            run_as_user='artys',
            bash_command=(
                'cd /home/artys/artys_ndmi_retriever; '
                f'sudo /home/artys/py373venv/bin/python /home/artys/artys_ndmi_retriever/EO_manager.pyc -f /home/artys/artys_ndmi_retriever/training.geojson -m hist -s {start_date} -e {end_date} -o /home/artys/artys_ndmi_retriever/latest -x 256 -y 256')
            )
        artys_ndmi_retriever.execute({})

    @task( task_id = "check_new_ndmi" )
    def check_new_ndmi() -> bool:
        """
        
        """
        from ndmi_prediction.transform import check_latest_ndmi
        # get latest ndmi retrieved by artys and move on main folder
        is_new = check_latest_ndmi() # if found latest ndmi then move to latest folder
        return is_new
    
    @task.branch( task_id = "branch_if_new_ndmi" )
    def ndmi_handler(is_new: bool) -> str:
        """
        
        """        
        # if there is a new ndmi the DAG is branched invoking the corresponding task
        if is_new:
            return "save_latest_ndmi_hf"
        else:
            return "fetch_latest_ndmi"
    
    @task(task_id="save_latest_ndmi_hf")
    def save_latest_ndmi_hf() -> None:
        """Handle new ndmi: save on hf new ndmi
        
           @return boolean indicating if  a new NDMI image has been downloaded 
        """
                
        # if there is a new ndmi by artys then clean temp content
        delete_temp_latest = BashOperator(
            task_id="delete_temp_latest",
            run_as_user='artys',
            bash_command=('sudo rm -r /home/artys/artys_ndmi_retriever/latest/*; ' )
        )
        delete_temp_latest.execute({})
            
        #save tif on huggingface
        from ndmi_prediction.load import save_ndmi_on_hf
        save_ndmi_on_hf(AIRFLOW_NDMI_FOLDER, "TMQ", Variable.get("HF_NDMI_DS"))

    @task(
            task_id="fetch_latest_ndmi",
            trigger_rule=TriggerRule.NONE_FAILED_MIN_ONE_SUCCESS)
    def fetch_latest_ndmi() -> str:
        """
        Handle ndmi: retrieve last ndmi searching both in 
        the local filesystem or in the repository
        """
        # check if there is already a latest ndmi in main folder
        from ndmi_prediction.extract import get_latest_ndmi
        local_file_path = get_latest_ndmi()
        #return both filesystem local path of the NDMI and boolean variable to indicate whether it is a new NDMI
        return local_file_path
    
    @task
    def pluvio_retriever() -> dict:
        print("Started pluvio retriever")
        from ndmi_prediction.extract import get_pluvio_from_ftp_server
        now = datetime.now()
        current_date = now.strftime("%Y%m%dT%H")+"00"
        latest_hour = now - timedelta(hours=1)
        local_paths = get_pluvio_from_ftp_server(latest_hour)
        result = {
            'paths': local_paths,
            'current_date': current_date
        }
        return result
    
    @task
    def compute_hourly_pluvio(latest_pluvio_local_paths: List[str], current_date: str) -> str:
        print("Computing latest hourly pluvio image")
        from ndmi_prediction.transform import compute_hourly_pluvio
        return compute_hourly_pluvio(latest_pluvio_local_paths, current_date)

    @task
    def update_cumulated(is_new_ndmi: bool, hourly_pluvio_filepath: str) -> str:  
        from ndmi_prediction.transform import reset_cumulated_pluvio, update_cumulated_pluvio
        if is_new_ndmi:
            cumulated_pluvio_file_path = reset_cumulated_pluvio(hourly_pluvio_filepath)
        else:
            cumulated_pluvio_file_path =  update_cumulated_pluvio(hourly_pluvio_filepath)

        delete_ftp_output_files = BashOperator(
            task_id="delete_ftp_output_files",
            run_as_user='artys',
            bash_command=('sudo rm -r /home/artys/ftp_output/*; ' )
        )
        delete_ftp_output_files.execute({}) #clean ftp_output files
        return cumulated_pluvio_file_path

    @task
    def retrieve_mlm_ndmi() -> str:
        from ndmi_prediction.extract import get_mlm_ndmi_from_hf
        model_local_path = get_mlm_ndmi_from_hf()
        return model_local_path

    @task
    def create_new_predicted_ndmi(cumulated_pluvio_path, latest_ndmi_path, model_path ):
        from ndmi_prediction.transform import run_mlm_ndmi
        from ndmi_prediction.load import save_ndmi_on_hf
        timestamp = datetime.now().strftime('%Y%m%dT%H') + "00" #name of the file and current timestamp
        run_mlm_ndmi(timestamp) #create new predicted ndmi
        save_ndmi_on_hf(AIRFLOW_PREDICTED_NDMI_FOLDER, timestamp, Variable.get("HF_NDMI_PREDICTION_DS")) #save new predicted ndmi on hf
        delete_predicted_ndmi = BashOperator(
            task_id="delete_predicted_ndmi",
            run_as_user='airflow',
            bash_command=('sudo rm -r /home/airflow/predicted_ndmi/*; ' )
        )
        delete_predicted_ndmi.execute({}) #remove local predicted ndmi files


    is_new_ndmi = check_new_ndmi()
    ndmi_handler = ndmi_handler(is_new_ndmi)
    fetch_latest_ndmi = fetch_latest_ndmi()
    
    
    model_local_path = retrieve_mlm_ndmi() #fase 1
    result = pluvio_retriever() #fase 1

    artys_ndmi_retriever() >> is_new_ndmi >> ndmi_handler #fase 1
    ndmi_handler >> save_latest_ndmi_hf() >> fetch_latest_ndmi #fase 2 A
    ndmi_handler >> fetch_latest_ndmi #fase 2 B

    hourly_pluvio_filepath = compute_hourly_pluvio(result['paths'], result['current_date'])
    cumulated_path = update_cumulated(is_new_ndmi, hourly_pluvio_filepath) #fase 2

    create_new_predicted_ndmi(cumulated_path, fetch_latest_ndmi, model_local_path) #fase 3

NDMIPrediction()
