import os
from datetime import datetime, timedelta
from typing import Any, List
from ndmi_prediction.config import AIRFLOW_MODELS_FOLDER, NDMI_MODEL_NAME

def get_pluvio_from_ftp_server(date: datetime) -> List[str]:
    import paramiko
    from paramiko.ssh_exception import SSHException
    from airflow.models import Variable
    # Get Variables from Airflow compose configuration
    hostname = Variable.get("FTP_HOSTNAME")
    secure_port = 22
    username = Variable.get("FTP_USERNAME")
    password = Variable.get("FTP_PASSWORD")
    base_dir = Variable.get("FTP_BASE_DIR")
    date_format = Variable.get("FTP_DATE_FORMAT")
    output_dir = Variable.get("FTP_OUTPUT_DIR")
    file_prefix = Variable.get("FTP_FILE_PREFIX")
    transport = None
    sftp = None
    try:
        # Set up the transport connection
        transport = paramiko.Transport((hostname, secure_port))
        transport.connect(None, username, password)
        # Create the SFTP client
        sftp = paramiko.SFTPClient.from_transport(transport)
        # Each file is in a folder specific to a date "YYYYMMDD"
        formatted_date = date.strftime(date_format)
        # Create a copy of the current date to edit it while keeping the right date
        current_date = _copy_date(date)
        # Array of all paths of pluvio new data 
        latest_pluvio_data_paths = []
        # Loop over until the hour changes
        while current_date.hour == date.hour:
            # Format the current date
            formatted_current_date = current_date.strftime("%Y%m%dT%H%M")
            filename = f"{file_prefix}_{formatted_current_date}.tif"
            file_path = os.path.join(base_dir, formatted_date, filename)
            local_path = os.path.join(output_dir, filename)
            print("LOCAL PATH", local_path)
            print("FILE PATH", file_path)
            # Get the file
            sftp.get(file_path, local_path)
            # append new file just downloaded to latest_pluvio_data_paths variable
            latest_pluvio_data_paths.append(local_path)
            # Increase the date by 5 minutes
            current_date = current_date + timedelta(minutes=5)
        if sftp:
            sftp.close()
        if transport:
            transport.close()
        # return filesystem local path where the file has been stored
        return latest_pluvio_data_paths
    except SSHException as e:
        print("An SSHException occurred")
        raise e
    finally:
        if sftp:
            print("Closing SFTP client")
            sftp.close()
        if transport:
            print("Closing FTP transport")
            transport.close()

def get_latest_ndmi() -> str:
    from ndmi_prediction.config import NDMI_DEST_FOLDER

    # check if there is already a latest ndmi in main folder
    search_kw = ".tif"
    local_file_path = check_local_file_exist(NDMI_DEST_FOLDER, search_kw)

    # if there isn't then retrieve latest ndmi from hugging face
    if not local_file_path:
        from ndmi_prediction.extract import get_latest_ndmi_from_hf
        local_file_path = get_latest_ndmi_from_hf()
    return local_file_path

def get_latest_ndmi_from_hf() -> str:
    from datasets import load_dataset, DownloadMode
    from airflow.models import Variable
    from ndmi_prediction.config import NDMI_DEST_FOLDER, AUTH_TOKEN

    HF_DATASET_NDMI = Variable.get("HF_NDMI_DS")
    
    ndmi_dataset = load_dataset(HF_DATASET_NDMI, 'latest', use_auth_token=AUTH_TOKEN,
                       download_mode=DownloadMode.FORCE_REDOWNLOAD, split='train')
   
    name = "latest_ndmi"
    ndmi_file_path = _save_tif_img(ndmi_dataset, NDMI_DEST_FOLDER, name)
    # return filesystem local path where the image has been stored
    return ndmi_file_path

def get_latest_cum_pluvio() -> str:
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER

    # create folder if not exists
    if not os.path.exists(PLUVIO_DEST_FOLDER):
        os.mkdir(PLUVIO_DEST_FOLDER)

    # check if there is already a latest cumulated pluvio stored locally
    search_kw = ".tif"
    local_file_path = check_local_file_exist(PLUVIO_DEST_FOLDER, search_kw)

    # if there isn't then retrieve latest ndmi from hugging face
    if not local_file_path:
        from ndmi_prediction.extract import get_latest_cum_pluvio_hf
        local_file_path = get_latest_cum_pluvio_hf()
    return local_file_path

def get_latest_cum_pluvio_hf() -> str:
    from airflow.models import Variable
    from datasets import load_dataset, DownloadMode 
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER, AUTH_TOKEN

    HF_H_CUM_PLUVIO_DATASET = Variable.get("HF_H_CUM_PLUVIO_DS")
    
    h_cum_pluvio = load_dataset(HF_H_CUM_PLUVIO_DATASET, name="latest", use_auth_token=AUTH_TOKEN,
                       download_mode=DownloadMode.FORCE_REDOWNLOAD, split='train')
    
    name = f"latest_pluvio"
    print("LOCAL FILENAME LATEST PLUVIO", name)
    cum_pluvio_file_path = _save_tif_img(h_cum_pluvio, PLUVIO_DEST_FOLDER, name)
    return cum_pluvio_file_path
    

def get_mlm_ndmi_from_hf() -> str:
    from huggingface_hub import hf_hub_download
    from airflow.models import Variable
    from ndmi_prediction.config import AUTH_TOKEN
    file_path = AIRFLOW_MODELS_FOLDER
    HF_DATASET_NDMI = Variable.get("HF_NDMI_DS")
    filename= NDMI_MODEL_NAME
    print("AUTH_TOKEN",AUTH_TOKEN)
    print("HF_DATASET_NDMI",HF_DATASET_NDMI)
    print("filename",filename)
    model_local_path = hf_hub_download(repo_id=HF_DATASET_NDMI, filename=filename, repo_type="model", local_dir=file_path, token=AUTH_TOKEN)
    # return filesystem local path where the model has been stored from the repository
    return model_local_path
    
# Utility functions

def check_local_file_exist(dest_folder: str, search_kw: str) -> str:
    import os
    file_list = os.listdir(dest_folder)

    for file_name in file_list:
        if file_name.endswith(search_kw):
            return os.path.join(dest_folder, file_name)  
    return str()

def _save_tif_img(image: Any, dest_folder: str, name: str) -> str:
    import rasterio
    import numpy as np

    current_img = image[0]
    current_img["image"] = np.array(current_img["image"])
    # Get the metadata
    img_dtype = current_img["dtype"] if "dtype" in image else rasterio.float32
    img_transform = current_img["transform"]
    img_crs = current_img["crs"]
    # Create dtype and crs instances
    image_dtype = np.dtype(img_dtype)
    image_crs = rasterio.crs.CRS.from_string(img_crs)
 
    filepath = os.path.join(dest_folder, name)
    with rasterio.open(f"{filepath}.tif", 'w',
        width=256, height=256,
        dtype=image_dtype,
        transform=img_transform,
        crs=image_crs,
        count=1) as r:
            r.write(current_img["image"])
            r.update_tags(
                timestamp=current_img["timestamp"].strftime('%Y%m%dT%H%M'),
                AREA_OR_POINT='Area'
            )
    # return filesystem local path where the image has been stored
    return os.path.join(dest_folder, f"{name}.tif")

def _copy_date(date: datetime) -> datetime:
    return datetime(day=date.day, month=date.month, year=date.year, hour=date.hour)
