from ndmi_prediction.config import AIRFLOW_NDMI_FOLDER, ARTYS_LATEST_NDMI_FOLDER, AIRFLOW_MODELS_FOLDER, NDMI_MODEL_NAME, AIRFLOW_CUMULATED_PLUVIO_FOLDER, AIRFLOW_PREDICTED_NDMI_FOLDER

import datetime
from typing import List


def check_latest_ndmi():
    import os
    import rasterio
    import numpy as np
    from airflow.operators.bash import BashOperator
    from ndmi_prediction.config import NDMI_DEST_FOLDER
    from ndmi_prediction.extract import check_local_file_exist
    isNew = False
    source_folder = ARTYS_LATEST_NDMI_FOLDER
    destination_folder = AIRFLOW_NDMI_FOLDER
    search_keyword = "TMQ"

    if not os.path.exists(NDMI_DEST_FOLDER):
        os.mkdir(NDMI_DEST_FOLDER)

    file_list = os.listdir(source_folder)

    for file_name in file_list:
        if search_keyword in file_name:
            if check_local_file_exist(NDMI_DEST_FOLDER, search_keyword):
                clean_latest_folder = BashOperator(
                    task_id="clean_latest_folder",
                    run_as_user='artys',
                    bash_command=(f'sudo rm -r {AIRFLOW_NDMI_FOLDER}/*; ' )
                )
                clean_latest_folder.execute({})
            source_path = os.path.join(source_folder, file_name)
            destination_path = os.path.join(NDMI_DEST_FOLDER, file_name)
            timestamp = file_name.split('_')[0]
            with rasterio.open(source_path, 'r') as src:
                image_data = src.read()
                image_data = np.array(image_data).reshape([1,256,256])
                with rasterio.open(f"{destination_path}", 'w',
                    width=256, height=256,
                    dtype=rasterio.float64,
                    transform=rasterio.transform.from_gcps(src.gcps[0]),
                    crs='EPSG:32632',
                    count=1) as r:
                        r.write(image_data)
                        r.update_tags(
                            timestamp=timestamp,
                        )
            isNew=True
    return isNew

def compute_hourly_pluvio(latest_pluvio_data_paths: List[str], current_date: str) -> str:
    import numpy as np
    import os
    from rasterio import open, float64, transform 
    from airflow.models import Variable

    file_prefix = Variable.get("FTP_FILE_PREFIX")
    output_dir = Variable.get("FTP_OUTPUT_DIR")

    hourly_data = np.zeros((1, 256, 256))
    for path in latest_pluvio_data_paths:
        with open(path, 'r') as pl:
            new_data = pl.read()
            new_data = np.array(new_data).reshape([1,256,256])
            hourly_data = np.add(hourly_data, new_data)

    print("HOURLY DATA: ", str(hourly_data))
    latest_hourly_pluvio_filename = f"{current_date}.tif"
    latest_hourly_pluvio_filepath = os.path.join(output_dir, latest_hourly_pluvio_filename)
    print("LATEST HOURLY PLUVIO PATH ", latest_hourly_pluvio_filepath)
    with open(f"{latest_hourly_pluvio_filepath}", 'w', 
                width=256, height=256,
                dtype=float64,
                crs='EPSG:32632',
                count=1) as h_pl:
                    h_pl.write(hourly_data)
                    h_pl.update_tags(
                        timestamp=current_date
                    )
    return latest_hourly_pluvio_filepath
       
def update_cumulated_pluvio(hourly_pluvio_filepath: str) -> str:
    from ndmi_prediction.extract import get_latest_cum_pluvio
    from ndmi_prediction.load import save_latest_cumulated_pluvio_hf
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER
    import rasterio
    import numpy as np, os

    cum_pluvio_file_path = get_latest_cum_pluvio()

    latest_h_pluvio_filename = os.path.basename(hourly_pluvio_filepath)
    latest_h_pluvio_timestamp = latest_h_pluvio_filename.split('.')[0]

    

    # read the tif of the latest hourly pluvio downloaded from the ftp server
    with rasterio.open(hourly_pluvio_filepath, 'r') as h_pl:
        h_pluvio = h_pl.read()
        h_pluvio = np.array(h_pluvio).reshape([1,256,256])

    # read the tif of the latest cumulated pluvio and create latest_cum_pluvio as the sum between hourly pluvio and old cumulated pluvio
    with rasterio.open(cum_pluvio_file_path, 'r') as cum:
        old_cum_pluvio = cum.read()
        old_cum_pluvio = np.array(old_cum_pluvio).reshape([1,256,256])
        latest_cum_pluvio = np.add(old_cum_pluvio, h_pluvio)
        
        with rasterio.open(f"{cum_pluvio_file_path}", 'w', 
            width=256, height=256,
            dtype=rasterio.float64,
            transform=rasterio.transform.from_gcps(cum.gcps[0]),
            crs='EPSG:32632',
            count=1) as update:
                update.write(latest_cum_pluvio)
                update.update_tags(
                    timestamp = latest_h_pluvio_timestamp
                )

    _update_repo_config_file(latest_h_pluvio_timestamp)
    # save the last update on HuggingFace repository
    save_latest_cumulated_pluvio_hf(cum_pluvio_file_path, latest_h_pluvio_filename)
    
    # return the local path of the computed cumulated pluvio as input to other tasks 
    print("LATEST CUM PLUVIO PATH ", cum_pluvio_file_path)
    return cum_pluvio_file_path
            


def reset_cumulated_pluvio(latest_hourly_pluvio: str) -> str:
    import os, shutil
    from ndmi_prediction.load import save_latest_cumulated_pluvio_hf
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER

    # move last computed cumulated pluvio to training dataset for further use
    _move_cumulated_pluvio_to_training()
    # remove all files in the repository that contains partially cumulated pluvio 
    # clean content of repo's config file 
    _remove_hourly_cumulated_pluvio_hf()
    # save latest hourly pluvio data on HF repository downloaded from ARTYS ftp server 
    new_pluvio_filename = os.path.basename(latest_hourly_pluvio)
    new_config_timestamp = new_pluvio_filename.split('.')[0]
    _update_repo_config_file(new_config_timestamp)
    save_latest_cumulated_pluvio_hf(latest_hourly_pluvio)
    # move latest hourly Pluvio data to the destination folder "/home/airflow/pluvio"
    new_pluvio_local_path = os.path.join(PLUVIO_DEST_FOLDER, "latest_pluvio.tif")
    shutil.move(latest_hourly_pluvio, new_pluvio_local_path)
    return new_pluvio_local_path


def _move_cumulated_pluvio_to_training() -> None:
    import os
    from airflow.models import Variable
    from huggingface_hub import HfApi
    from ndmi_prediction.extract import get_latest_cum_pluvio
    from datetime import datetime
    from ndmi_prediction.load import update_config_json_file_on_hf
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER, AUTH_TOKEN
    HF_CUM_PLUVIO_TR_DATASET = Variable.get("HF_CUM_PLUVIO_TR_DS")
    
    hf_cumulated_api = HfApi( token=AUTH_TOKEN )

    cum_pluvio_file_path = get_latest_cum_pluvio()
    filename = datetime.now().strftime('%Y%m%d')
    print("FILENAME: ", filename)
    print("LOCAL FILE PATH: ", cum_pluvio_file_path)
    hf_cumulated_api.upload_file(
        path_or_fileobj=f"{cum_pluvio_file_path}",
        path_in_repo=f"{filename}",
        repo_id=HF_CUM_PLUVIO_TR_DATASET,
        repo_type='dataset'
    )

    update_config_json_file_on_hf(f"{filename}.tif", "filelist.json", HF_CUM_PLUVIO_TR_DATASET, PLUVIO_DEST_FOLDER) #update filelist.json with new date config


def _remove_hourly_cumulated_pluvio_hf() -> None:
    import json, os
    from huggingface_hub import hf_hub_download
    from airflow.models import Variable
    from huggingface_hub import HfApi
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER, AUTH_TOKEN

    HF_H_CUM_PLUVIO_DATASET = Variable.get("HF_H_CUM_PLUVIO_DS")
    hf_api = HfApi( token=AUTH_TOKEN )

    filename = "filelist.py"
    config_filepath = os.path.join(PLUVIO_DEST_FOLDER, filename)
    hf_hub_download(repo_id=HF_H_CUM_PLUVIO_DATASET, filename=filename, repo_type="dataset", local_dir=PLUVIO_DEST_FOLDER, token=AUTH_TOKEN)

    with open(config_filepath, 'r') as py_file:
        configs = py_file.readlines()[0]
        configs = configs.replace("'", '"')
        configs = configs.split(" = ")
        configs = json.loads(configs[1])
        
    
    # clean all files in the repository
    for conf in configs: 
        print("CONF: ", conf)
        hf_api.delete_file(path_in_repo=f"{conf}.tif" , repo_id=HF_H_CUM_PLUVIO_DATASET, token=AUTH_TOKEN, repo_type="dataset")
    
    # delete content of the configuration file and upload it back to the repo
    configs = r"CONFIGS = []"
    with open(config_filepath, 'w') as py_file:
        py_file.write(configs)

    hf_api.upload_file(
        path_or_fileobj=f"{config_filepath}",
        path_in_repo=f"{filename}",
        repo_id=HF_H_CUM_PLUVIO_DATASET,
        repo_type='dataset'
    )
        
def _update_repo_config_file(new_config_timestamp: str) -> None:
    import json, os
    from huggingface_hub import hf_hub_download
    from airflow.models import Variable
    from huggingface_hub import HfApi
    from ndmi_prediction.config import PLUVIO_DEST_FOLDER
    from ndmi_prediction.load import update_config_py_file_on_hf

    HF_H_CUM_PLUVIO_DATASET = Variable.get("HF_H_CUM_PLUVIO_DS")
    filename = "filelist.py"
    update_config_py_file_on_hf(new_config_timestamp, filename, HF_H_CUM_PLUVIO_DATASET, PLUVIO_DEST_FOLDER)
    

def get_latest_ndmi_filename():
    import os
    destination_folder = AIRFLOW_NDMI_FOLDER
    search_keyword = ".tif"
    name = ""
    file_list = os.listdir(destination_folder)

    for file_name in file_list:
        if search_keyword in file_name:
            name=file_name
    return name

def get_latest_cumulated_pluvio_filename():
    import os
    destination_folder = AIRFLOW_CUMULATED_PLUVIO_FOLDER
    search_keyword = ".tif"
    name = ""
    file_list = os.listdir(destination_folder)

    for file_name in file_list:
        if search_keyword in file_name:
            name=file_name
    return name

def gradient_alg(img): 
    import numpy as np
    img = img.reshape(256, 256)
    for i, row in enumerate(img):
        mask = np.isnan(row)

        x = np.arange(len(row))
        y = row[~mask]
        if (len(y) > 1):
            gradient = np.gradient(y)
            row[mask] = np.interp(x[mask], x[~mask], gradient)
        elif (len(y) == 1):
            row[mask]= np.ones(255)
            print("not enough points to calculate the gradient: ", len(y))
        else: 
            row[mask]= np.ones(256)
            print("not enough points to calculate the gradient: ", len(y))
        img[i]=row
    img = img.reshape(1, 256, 256, 1)
    return img

def run_mlm_ndmi(timestamp):
    import os
    import rasterio
    import tensorflow as tf
    import numpy as np
    width=256
    height=256
    destination_folder = AIRFLOW_PREDICTED_NDMI_FOLDER

    #retrieve mlm model
    filepath = f"{AIRFLOW_MODELS_FOLDER}/{NDMI_MODEL_NAME}"
    model = tf.keras.models.load_model(filepath, compile=False)

    #retrieve cumulated pluvio
    pluvio_source_path =  f"{AIRFLOW_CUMULATED_PLUVIO_FOLDER}/{get_latest_cumulated_pluvio_filename()}"
    with rasterio.open(pluvio_source_path, 'r') as src:
        pluvio_image_data = src.read()
        pluvio_image_data = np.array(pluvio_image_data)
        pluvio_image_data= pluvio_image_data.reshape([1,256,256,1])
    
    #retrieve last ndmi
    ndmi_source_path =  f"{AIRFLOW_NDMI_FOLDER}/{get_latest_ndmi_filename()}"
    with rasterio.open(ndmi_source_path, 'r') as src:
        src_ndmi_timestamp = src.tags()['timestamp']
        ndmi_image_data = src.read()
        ndmi_image_data = np.array(ndmi_image_data)
        ndmi_image_data = gradient_alg(ndmi_image_data) #gradient the image to remove NaN values
        gcps=src.gcps[0]

    #create input tensor
    aux = np.zeros((1, width, height, 1))
    input_tensor = np.concatenate([ndmi_image_data, pluvio_image_data, aux], axis=3)

    #run mlm
    predictions = model.predict(input_tensor)
    
    predictions = predictions.reshape([1,256,256])

    #save temp result as new predicted ndmi
    if not os.path.exists(destination_folder):
        os.mkdir(destination_folder)
    with rasterio.open(f"{destination_folder}/{timestamp}.tif", 'w',
        width=256, height=256,
        dtype=rasterio.float64,
        transform=rasterio.transform.from_gcps(gcps),
        crs='EPSG:32632',
        count=1) as r:
            r.write(predictions)
            r.update_tags(
                timestamp=timestamp,
                ndmi_timestamp=src_ndmi_timestamp,
                AREA_OR_POINT='Area'
            )

    


